@@ -48,6 +48,10 @@ from .extract_features import worker_read_split_holebatches_to_queue
 from .extract_features import worker_extract_features_from_holebatches
 from .extract_features import _get_holes
 
+# add this export temporarily
+# https://github.com/pytorch/pytorch/issues/37377
+os.environ['MKL_THREADING_LAYER'] = 'GNU'
+
 queue_size_border = 1000
 time_wait = 1
 
@@ -213,7 +217,7 @@ def _format_features_from_strbatch2s(featurestrs_batch_q, features_batch_q, seq_
 
 
 # call mods =============================================================
-def _call_mods2s(features_batch, model, batch_size):
+def _call_mods2s(features_batch, model, batch_size, device=0):
     sampleinfo, fkmers, fpasss, fipdms, fipdsds, fpwms, fpwsds, fquals, fmaps, \
         rkmers, rpasss, ripdms, ripdsds, rpwms, rpwsds, rquals, rmaps, labels = features_batch
     labels = np.reshape(labels, (len(labels)))
@@ -225,32 +229,34 @@ def _call_mods2s(features_batch, model, batch_size):
         batch_s, batch_e = i, i + batch_size
         b_sampleinfo = sampleinfo[batch_s:batch_e]
 
-        b_fkmers = fkmers[batch_s:batch_e]
-        b_fpasss = fpasss[batch_s:batch_e]
-        b_fipdms = fipdms[batch_s:batch_e]
-        b_fipdsds = fipdsds[batch_s:batch_e]
-        b_fpwms = fpwms[batch_s:batch_e]
-        b_fpwsds = fpwsds[batch_s:batch_e]
-        b_fquals = fquals[batch_s:batch_e]
-        b_fmaps = fmaps[batch_s:batch_e]
+        b_fkmers = np.array(fkmers[batch_s:batch_e])
+        b_fpasss = np.array(fpasss[batch_s:batch_e])
+        b_fipdms = np.array(fipdms[batch_s:batch_e])
+        b_fipdsds = np.array(fipdsds[batch_s:batch_e])
+        b_fpwms = np.array(fpwms[batch_s:batch_e])
+        b_fpwsds = np.array(fpwsds[batch_s:batch_e])
+        b_fquals = np.array(fquals[batch_s:batch_e])
+        b_fmaps = np.array(fmaps[batch_s:batch_e])
 
-        b_rkmers = rkmers[batch_s:batch_e]
-        b_rpasss = rpasss[batch_s:batch_e]
-        b_ripdms = ripdms[batch_s:batch_e]
-        b_ripdsds = ripdsds[batch_s:batch_e]
-        b_rpwms = rpwms[batch_s:batch_e]
-        b_rpwsds = rpwsds[batch_s:batch_e]
-        b_rquals = rquals[batch_s:batch_e]
-        b_rmaps = rmaps[batch_s:batch_e]
+        b_rkmers = np.array(rkmers[batch_s:batch_e])
+        b_rpasss = np.array(rpasss[batch_s:batch_e])
+        b_ripdms = np.array(ripdms[batch_s:batch_e])
+        b_ripdsds = np.array(ripdsds[batch_s:batch_e])
+        b_rpwms = np.array(rpwms[batch_s:batch_e])
+        b_rpwsds = np.array(rpwsds[batch_s:batch_e])
+        b_rquals = np.array(rquals[batch_s:batch_e])
+        b_rmaps = np.array(rmaps[batch_s:batch_e])
 
-        b_labels = labels[batch_s:batch_e]
+        b_labels = np.array(labels[batch_s:batch_e])
         if len(b_sampleinfo) > 0:
-            voutputs, vlogits = model(FloatTensor(b_fkmers), FloatTensor(b_fpasss), FloatTensor(b_fipdms),
-                                      FloatTensor(b_fipdsds), FloatTensor(b_fpwms), FloatTensor(b_fpwsds),
-                                      FloatTensor(b_fquals), FloatTensor(b_fmaps),
-                                      FloatTensor(b_rkmers), FloatTensor(b_rpasss), FloatTensor(b_ripdms),
-                                      FloatTensor(b_ripdsds), FloatTensor(b_rpwms), FloatTensor(b_rpwsds),
-                                      FloatTensor(b_rquals), FloatTensor(b_rmaps))
+            voutputs, vlogits = model(FloatTensor(b_fkmers, device), FloatTensor(b_fpasss, device),
+                                      FloatTensor(b_fipdms, device), FloatTensor(b_fipdsds, device),
+                                      FloatTensor(b_fpwms, device), FloatTensor(b_fpwsds, device),
+                                      FloatTensor(b_fquals, device), FloatTensor(b_fmaps, device),
+                                      FloatTensor(b_rkmers, device), FloatTensor(b_rpasss, device),
+                                      FloatTensor(b_ripdms, device), FloatTensor(b_ripdsds, device),
+                                      FloatTensor(b_rpwms, device), FloatTensor(b_rpwsds, device),
+                                      FloatTensor(b_rquals, device), FloatTensor(b_rmaps, device))
             _, vpredicted = torch.max(vlogits.data, 1)
             if use_cuda:
                 vlogits = vlogits.cpu()
@@ -267,7 +273,7 @@ def _call_mods2s(features_batch, model, batch_size):
                 # chromosome, pos, strand, holeid, loc, depth, prob_0, prob_1, called_label, seq
                 prob_0, prob_1 = logits[idx][0], logits[idx][1]
                 prob_0_norm = round(prob_0 / (prob_0 + prob_1), 6)
-                prob_1_norm = round(prob_1 / (prob_0 + prob_1), 6)
+                prob_1_norm = round(1 - prob_0_norm, 6)
                 b_idx_kmer = ''.join([code2base_dna[x] for x in b_fkmers[idx]])
                 center_idx = int(np.floor(len(b_idx_kmer)/2))
                 bkmer_start = center_idx - 2 if center_idx - 2 >= 0 else 0
@@ -282,7 +288,7 @@ def _call_mods2s(features_batch, model, batch_size):
     return pred_str, accuracy, batch_num
 
 
-def _call_mods_q(model_path, features_batch_q, pred_str_q, args):
+def _call_mods_q(model_path, features_batch_q, pred_str_q, args, device=0):
     print('call_mods process-{} starts'.format(os.getpid()))
     if args.model_type in {"attbigru2s", "attbilstm2s"}:
         model = ModelAttRNN(args.seq_len, args.layer_rnn, args.class_num,
@@ -292,20 +298,33 @@ def _call_mods_q(model_path, features_batch_q, pred_str_q, args):
                             is_map=str2bool(args.is_map),
                             is_stds=str2bool(args.is_stds),
                             is_npass=str2bool(args.is_npass),
-                            model_type=args.model_type)
+                            model_type=args.model_type,
+                            device=device)
     else:
         raise ValueError("--model_type not right!")
 
-    if use_cuda:
-        model = model.cuda()
-        para_dict = torch.load(model_path)
-    else:
+    try:
         para_dict = torch.load(model_path, map_location=torch.device('cpu'))
+        # para_dict = torch.load(model_path, map_location=torch.device(device))
+        model_dict = model.state_dict()
+        model_dict.update(para_dict)
+        model.load_state_dict(model_dict)
+        if str2bool(args.loginfo):
+            print('call_mods process-{} loads model param successfully'.format(os.getpid()))
+    except RuntimeError:
+        # for DDP model convertion (key: module.embed.weight -> embed.weight)
+        from collections import OrderedDict
+        para_dict = torch.load(model_path, map_location=torch.device('cpu'))
+        para_dict_new = OrderedDict()
+        for param_tensor in para_dict.keys():
+            para_dict_new[param_tensor[7:]] = para_dict[param_tensor]
+        model.load_state_dict(para_dict_new)
+        if str2bool(args.loginfo):
+            print('call_mods process-{} loads model param successfully-1'.format(os.getpid()))
+    sys.stdout.flush()
 
-    model_dict = model.state_dict()
-    model_dict.update(para_dict)
-    model.load_state_dict(model_dict)
-
+    if use_cuda:
+        model = model.cuda(device)
     model.eval()
 
     accuracy_list = []
@@ -322,7 +341,7 @@ def _call_mods_q(model_path, features_batch_q, pred_str_q, args):
             break
 
         if args.model_type in {"attbigru2s", "attbilstm2s"}:
-            pred_str, accuracy, batch_num = _call_mods2s(features_batch, model, args.batch_size)
+            pred_str, accuracy, batch_num = _call_mods2s(features_batch, model, args.batch_size, device)
         else:
             raise ValueError("--model_type not right!")
 
@@ -362,6 +381,15 @@ def _write_predstr_to_file(write_fp, predstr_q, is_gzip):
         wf.flush()
 
 
+def _get_gpus():
+    num_gpus = torch.cuda.device_count()
+    if num_gpus > 0:
+        gpulist = list(range(num_gpus))
+    else:
+        gpulist = [0]
+    return gpulist * 1000
+
+
 def call_mods(args):
     print("[main]call_mods starts..")
     start = time.time()
@@ -394,9 +422,9 @@ def call_mods(args):
         index_bam_if_needed2(input_path, args.threads)
 
         dnacontigs = None
-        if args.mode == "reference":
+        if args.mode == "align":
             if args.ref is None:
-                raise ValueError("--ref must be provided when using reference mode!")
+                raise ValueError("--ref must be provided when using align mode!")
             reference = os.path.abspath(args.ref)
             if not os.path.exists(reference):
                 raise IOError("refernce(--ref) file does not exist!")
@@ -416,15 +444,17 @@ def call_mods(args):
         else:
             if nproc_dp > nproc_to_call_mods_in_cpu_mode:
                 nproc_dp = nproc_to_call_mods_in_cpu_mode
-        if nproc <= nproc_dp + 2:
-            print("--threads must be > nproc_dp + 2!!")
-            nproc = nproc_dp + 2 + 1
+        if nproc <= nproc_dp + 3:
+            print("--threads must be > --threads_call + 3!!")
+            nproc = nproc_dp + 3 + 1  # 2 for reading, 1 for writing, 1 for extracting
 
+        # 2 threads/processes
         p_read = mp.Process(target=worker_read_split_holebatches_to_queue,
-                            args=(input_path, hole_batch_q, args))
+                            args=(input_path, hole_batch_q, 2, args))
         p_read.daemon = True
         p_read.start()
 
+        # 1 processes
         p_w = mp.Process(target=_write_predstr_to_file, args=(out_per_readsite, pred_str_q, args.gzip))
         p_w.daemon = True
         p_w.start()
@@ -432,7 +462,9 @@ def call_mods(args):
         # TODO: why the processes in ps_extract start so slowly?
         ps_extract = []
         ps_call = []
-        nproc_ext = nproc - nproc_dp - 2
+        nproc_ext = nproc - nproc_dp - 3
+        gpulist = _get_gpus()
+        gpuindex = 0
         for i in range(max(nproc_ext, nproc_dp)):
             if i < nproc_ext:
                 p = mp.Process(target=worker_extract_features_from_holebatches,
@@ -443,7 +475,9 @@ def call_mods(args):
                 p.start()
                 ps_extract.append(p)
             if i < nproc_dp:
-                p = mp.Process(target=_call_mods_q, args=(model_path, features_batch_q, pred_str_q, args))
+                p = mp.Process(target=_call_mods_q, args=(model_path, features_batch_q, pred_str_q,
+                                                          args, gpulist[gpuindex]))
+                gpuindex += 1
                 p.daemon = True
                 p.start()
                 ps_call.append(p)
@@ -464,7 +498,8 @@ def call_mods(args):
             from ._bam2modbam import add_mm_ml_tags_to_bam
             out_modbam = args.output + ".modbam.bam"
             add_mm_ml_tags_to_bam(input_path, out_per_readsite, out_modbam,
-                                  rm_pulse=True, threads=args.threads)
+                                  rm_pulse=True, threads=args.threads,
+                                  mode=args.mode)
     else:
         # features_batch_q = mp.Queue()
         features_batch_q = Queue()
@@ -481,8 +516,8 @@ def call_mods(args):
             if nproc_dp > nproc_to_call_mods_in_cpu_mode:
                 nproc_dp = nproc_to_call_mods_in_cpu_mode
         if nproc < nproc_dp + 2:
-            print("--threads must be > nproc_dp + 2!!")
-            nproc = nproc_dp + 2 + 1
+            print("--threads must be > --threads_call + 2!!")
+            nproc = nproc_dp + 2 + 1  # 1 for reading, 1 for writing, 1 for extracting
         nproc_cnvt = nproc - nproc_dp - 2
 
         p_read = mp.Process(target=_read_features_file_to_str, args=(input_path, featurestrs_batch_q,
@@ -504,8 +539,12 @@ def call_mods(args):
             raise ValueError("--model_type not right!")
 
         predstr_procs = []
+        gpulist = _get_gpus()
+        gpuindex = 0
         for _ in range(nproc_dp):
-            p = mp.Process(target=_call_mods_q, args=(model_path, features_batch_q, pred_str_q, args))
+            p = mp.Process(target=_call_mods_q, args=(model_path, features_batch_q, pred_str_q,
+                                                      args, gpulist[gpuindex]))
+            gpuindex += 1
             p.daemon = True
             p.start()
             predstr_procs.append(p)
@@ -641,9 +680,9 @@ def main():
 
     parser.add_argument("--threads", "-p", action="store", type=int, default=10,
                         required=False, help="number of threads to be used, default 10.")
-    parser.add_argument("--threads_call", action="store", type=int, default=2,
-                        required=False, help="number of threads used to call with trained models, "
-                                             "no more than threads/4 is suggested. default 2.")
+    parser.add_argument("--threads_call", action="store", type=int, default=3,
+                        required=False, help="number of threads used to call modifications with trained "
+                                             "models, no more than threads/3 is suggested. default 3.")
     parser.add_argument('--tseed', type=int, default=1234,
                         help='random seed for torch')
 
