@@ -120,6 +120,8 @@ def _get_moddict(readitem, modbase="C", modification="m"):
 
     # use .modified_bases instead of MM/ML tags to get moddict
     modinfo = readitem.modified_bases
+    if modinfo is None:
+        return {}
     modtuple = None
     for modkey in modinfo.keys():
         if modkey[0] == modbase and modkey[2] == modification:
@@ -162,6 +164,8 @@ def _call_modfreq_of_one_region(refpos2modinfo, args):
             info_hp2 = _cal_modprob_in_count_mode(hp2_mods, args.prob_cf) if len(hp2_mods) > 0 else None
             refpos_results.append((refpos, info_all, info_hp1, info_hp2))
     elif args.call_mode == "aggregate":
+        import torch
+        from .models import AggrAttRNN
         pass
 
     return refpos_results
@@ -461,6 +465,8 @@ def main():
     scfb_callfreq.add_argument('--call_mode', type=str, action="store", required=False, default="count",
                                choices=["count", "aggregate"],
                                help='call mode: count, aggregate. default count.')
+    scfb_callfreq.add_argument("--ag_model", type=str, action="store", required=False,
+                               help='model path for call_mode aggregate')
     scfb_callfreq.add_argument('--prob_cf', type=float, action="store", required=False, default=0.0,
                                help='this is to remove ambiguous calls. '
                                'if abs(prob1-prob0)>=prob_cf, then we use the call. e.g., proc_cf=0 '
