@@ -58,10 +58,23 @@ def main_train(args):
     train(args)
 
 
+def main_trainm(args):
+    from .train_multigpu import train
+
+    display_args(args)
+    train(args)
+
+
+# def main_denoise(args):
+#     from .denoise import denoise
+#     display_args(args)
+#     denoise(args)
+
+
 def main():
     parser = argparse.ArgumentParser(prog='ccsmeth',
-                                     description="detecting methylation from PacBio CCS reads, "
-                                                 "ccsmeth contains 7 modules:\n"
+                                     description="ccsmeth detects methylation from PacBio CCS reads, "
+                                                 "which contains 8 modules:\n"
                                                  "\t%(prog)s call_hifi: call hifi reads from subreads "
                                                  "using CCS (PBCCS)\n"
                                                  "\t%(prog)s call_mods: call modifications\n"
@@ -73,7 +86,10 @@ def main():
                                                  "\t%(prog)s extract: extract features from hifi reads "
                                                  "for training or testing\n"
                                                  "\t%(prog)s train: train a model, need two independent "
-                                                 "datasets for training and validating",
+                                                 "datasets for training and validating\n"
+                                                 "\t%(prog)s trainm: [EXPERIMENTAL]train a model using multi "
+                                                 "gpus",
+                                                 # "\t%(prog)s denoise: [EXPERIMENTAL]denoise",
                                      formatter_class=argparse.RawTextHelpFormatter)
     parser.add_argument(
         '-v', '--version', action='version',
@@ -96,6 +112,8 @@ def main():
     sub_extract = subparsers.add_parser("extract", description="extract features from hifi reads.")
     sub_train = subparsers.add_parser("train", description="train a model, need two independent datasets for training "
                                                            "and validating")
+    sub_trainm = subparsers.add_parser("trainm", description="[EXPERIMENTAL]train a model using multi gpus")
+    # sub_denoise = subparsers.add_parser("denoise", description="[EXPERIMENTAL]denoise")
 
     # sub_call_hifi ============================================================================
     sch_input = sub_call_hifi.add_argument_group("INPUT")
@@ -290,21 +308,15 @@ def main():
 
     sub_call_mods.add_argument("--threads", "-p", action="store", type=int, default=10,
                                required=False, help="number of threads to be used, default 10.")
-    sub_call_mods.add_argument("--threads_call", action="store", type=int, default=2,
-                               required=False, help="number of threads used to call with trained models, "
-                                                    "no more than threads/4 is suggested. default 2.")
+    sub_call_mods.add_argument("--threads_call", action="store", type=int, default=3,
+                               required=False, help="number of threads used to call modifications with trained "
+                                                    "models, no more than threads/3 is suggested. default 3.")
     sub_call_mods.add_argument('--tseed', type=int, default=1234,
                                help='random seed for torch')
 
     sub_call_mods.set_defaults(func=main_call_mods)
 
     # sub_extract ============================================================================
-    sub_extract.add_argument("--threads", type=int, default=5, required=False,
-                             help="number of threads, default 5")
-    sub_extract.add_argument("--loginfo", type=str, default="no", required=False,
-                             help="if printing more info of feature extraction on reads. "
-                                  "yes or no, default no")
-
     se_input = sub_extract.add_argument_group("INPUT")
     se_input.add_argument("--input", "-i", type=str, required=True,
                           help="input file in bam/sam format, "
@@ -369,6 +381,12 @@ def main():
     se_extract_ref.add_argument("--skip_unmapped", type=str, default="yes", required=False,
                                 help="if skipping unmapped sites in reads, yes or no, default yes")
 
+    sub_extract.add_argument("--threads", type=int, default=5, required=False,
+                             help="number of threads, default 5")
+    sub_extract.add_argument("--loginfo", type=str, default="no", required=False,
+                             help="if printing more info of feature extraction on reads. "
+                                  "yes or no, default no")
+
     sub_extract.set_defaults(func=main_extract)
 
     # sub_call_freq_txt =====================================================================================
@@ -459,6 +477,8 @@ def main():
     scfb_callfreq.add_argument('--call_mode', type=str, action="store", required=False, default="count",
                                choices=["count", "aggregate"],
                                help='call mode: count, aggregate. default count.')
+    scfb_callfreq.add_argument("--ag_model", type=str, action="store", required=False,
+                               help='model path for call_mode aggregate')
     scfb_callfreq.add_argument('--prob_cf', type=float, action="store", required=False, default=0.0,
                                help='this is to remove ambiguous calls. '
                                     'if abs(prob1-prob0)>=prob_cf, then we use the call. e.g., proc_cf=0 '
@@ -500,7 +520,7 @@ def main():
     st_output = sub_train.add_argument_group("OUTPUT")
     st_output.add_argument('--model_dir', type=str, required=True)
 
-    st_train = sub_train.add_argument_group("TRAIN")
+    st_train = sub_train.add_argument_group("TRAIN MODEL_HYPER")
     # model param
     st_train.add_argument('--model_type', type=str, default="attbigru2s",
                           choices=["attbilstm2s", "attbigru2s"],
@@ -530,29 +550,222 @@ def main():
     st_train.add_argument('--hid_rnn', type=int, default=256, required=False,
                           help="BiRNN hidden_size for combined feature")
 
+    st_training = sub_train.add_argument_group("TRAINING")
     # model training
-    st_train.add_argument('--optim_type', type=str, default="Adam", choices=["Adam", "RMSprop", "SGD",
-                                                                             "Ranger"],
-                          required=False, help="type of optimizer to use, 'Adam' or 'SGD' or 'RMSprop' or 'Ranger', "
-                                               "default Adam")
-    st_train.add_argument('--batch_size', type=int, default=512, required=False)
-    st_train.add_argument('--lr', type=float, default=0.001, required=False)
-    st_train.add_argument('--lr_decay', type=float, default=0.1, required=False)
-    st_train.add_argument('--lr_decay_step', type=int, default=1, required=False)
-    st_train.add_argument("--max_epoch_num", action="store", default=50, type=int,
-                          required=False, help="max epoch num, default 50")
-    st_train.add_argument("--min_epoch_num", action="store", default=10, type=int,
-                          required=False, help="min epoch num, default 10")
-    st_train.add_argument('--pos_weight', type=float, default=1.0, required=False)
-    st_train.add_argument('--tseed', type=int, default=1234,
-                          help='random seed for pytorch')
-    st_train.add_argument('--step_interval', type=int, default=500, required=False)
+    st_training.add_argument('--optim_type', type=str, default="Adam", choices=["Adam", "RMSprop", "SGD",
+                                                                                "Ranger"],
+                             required=False, help="type of optimizer to use, 'Adam' or 'SGD' or 'RMSprop' "
+                                                  "or 'Ranger', default Adam")
+    st_training.add_argument('--batch_size', type=int, default=512, required=False)
+    st_training.add_argument('--lr_scheduler', type=str, default='StepLR', required=False,
+                             choices=["StepLR", "ReduceLROnPlateau"],
+                             help="StepLR or ReduceLROnPlateau, default StepLR")
+    st_training.add_argument('--lr', type=float, default=0.001, required=False,
+                             help="default 0.001")
+    st_training.add_argument('--lr_decay', type=float, default=0.1, required=False,
+                             help="default 0.1")
+    st_training.add_argument('--lr_decay_step', type=int, default=1, required=False,
+                             help="effective in StepLR. default 1")
+    st_training.add_argument('--lr_patience', type=int, default=0, required=False,
+                             help="effective in ReduceLROnPlateau. default 0")
+    st_training.add_argument('--lr_mode_strategy', type=str, default="last", required=False,
+                             choices=["last", "mean", "max"],
+                             help="effective in ReduceLROnPlateau. last, mean, or max, default last")
+    st_training.add_argument("--max_epoch_num", action="store", default=50, type=int,
+                             required=False, help="max epoch num, default 50")
+    st_training.add_argument("--min_epoch_num", action="store", default=10, type=int,
+                             required=False, help="min epoch num, default 10")
+    st_training.add_argument('--pos_weight', type=float, default=1.0, required=False)
+    st_training.add_argument('--step_interval', type=int, default=500, required=False)
+    st_training.add_argument('--dl_num_workers', type=int, default=0, required=False,
+                             help="default 0")
+    st_training.add_argument('--dl_offsets', action="store_true", default=False, required=False,
+                             help="use file offsets loader")
 
-    st_train.add_argument('--init_model', type=str, default=None, required=False,
-                          help="file path of pre-trained model parameters to load before training")
+    st_training.add_argument('--init_model', type=str, default=None, required=False,
+                             help="file path of pre-trained model parameters to load before training")
+    st_training.add_argument('--tseed', type=int, default=1234,
+                             help='random seed for pytorch')
 
     sub_train.set_defaults(func=main_train)
 
+    # sub_train_multigpu =====================================================================================
+    stm_input = sub_trainm.add_argument_group("INPUT")
+    stm_input.add_argument('--train_file', type=str, required=True)
+    stm_input.add_argument('--valid_file', type=str, required=True)
+
+    stm_input = sub_trainm.add_argument_group("OUTPUT")
+    stm_input.add_argument('--model_dir', type=str, required=True)
+
+    # model param
+    stm_train = sub_trainm.add_argument_group("TRAIN MODEL_HYPER")
+    stm_train.add_argument('--model_type', type=str, default="attbigru2s",
+                           choices=["attbilstm2s", "attbigru2s"],
+                           required=False,
+                           help="type of model to use, 'attbilstm2s', 'attbigru2s', "
+                                "default: attbigru2s")
+    stm_train.add_argument('--seq_len', type=int, default=21, required=False,
+                           help="len of kmer. default 21")
+    stm_train.add_argument('--is_npass', type=str, default="yes", required=False,
+                           help="if using num_pass features, yes or no, default yes")
+    stm_train.add_argument('--is_qual', type=str, default="no", required=False,
+                           help="if using base_quality features, yes or no, default no")
+    stm_train.add_argument('--is_map', type=str, default="no", required=False,
+                           help="if using mapping features, yes or no, default no")
+    stm_train.add_argument('--is_stds', type=str, default="no", required=False,
+                           help="if using std features, yes or no, default no")
+    stm_train.add_argument('--class_num', type=int, default=2, required=False)
+    stm_train.add_argument('--dropout_rate', type=float, default=0.5, required=False)
+
+    # BiRNN model param
+    stm_train.add_argument('--n_vocab', type=int, default=16, required=False,
+                           help="base_seq vocab_size (15 base kinds from iupac)")
+    stm_train.add_argument('--n_embed', type=int, default=4, required=False,
+                           help="base_seq embedding_size")
+    stm_train.add_argument('--layer_rnn', type=int, default=3,
+                           required=False, help="BiRNN layer num, default 3")
+    stm_train.add_argument('--hid_rnn', type=int, default=256, required=False,
+                           help="BiRNN hidden_size for combined feature")
+
+    # model training
+    stm_training = sub_trainm.add_argument_group("TRAINING")
+    stm_training.add_argument('--optim_type', type=str, default="Adam", choices=["Adam", "RMSprop", "SGD",
+                                                                                 "Ranger"],
+                              required=False, help="type of optimizer to use, 'Adam' or 'SGD' or 'RMSprop' "
+                                                   "or 'Ranger', default Adam")
+    stm_training.add_argument('--batch_size', type=int, default=512, required=False)
+    stm_training.add_argument('--lr_scheduler', type=str, default='StepLR', required=False,
+                              choices=["StepLR", "ReduceLROnPlateau"],
+                              help="StepLR or ReduceLROnPlateau, default StepLR")
+    stm_training.add_argument('--lr', type=float, default=0.001, required=False,
+                              help="default 0.001. [lr should be lr*world_size when using multi gpus? "
+                                   "or lower batch_size?]")
+    stm_training.add_argument('--lr_decay', type=float, default=0.1, required=False,
+                              help="default 0.1")
+    stm_training.add_argument('--lr_decay_step', type=int, default=1, required=False,
+                              help="effective in StepLR. default 1")
+    stm_training.add_argument('--lr_patience', type=int, default=0, required=False,
+                              help="effective in ReduceLROnPlateau. default 0")
+    # stm_training.add_argument('--lr_mode_strategy', type=str, default="last", required=False,
+    #                           choices=["last", "mean", "max"],
+    #                           help="effective in ReduceLROnPlateau. last, mean, or max, default last")
+    stm_training.add_argument("--max_epoch_num", action="store", default=50, type=int,
+                              required=False, help="max epoch num, default 50")
+    stm_training.add_argument("--min_epoch_num", action="store", default=10, type=int,
+                              required=False, help="min epoch num, default 10")
+    stm_training.add_argument('--pos_weight', type=float, default=1.0, required=False)
+    stm_training.add_argument('--step_interval', type=int, default=500, required=False)
+    stm_training.add_argument('--dl_num_workers', type=int, default=0, required=False,
+                              help="default 0")
+
+    stm_training.add_argument('--init_model', type=str, default=None, required=False,
+                              help="file path of pre-trained model parameters to load before training")
+    stm_training.add_argument('--tseed', type=int, default=1234,
+                              help='random seed for pytorch')
+
+    stm_trainingp = sub_trainm.add_argument_group("TRAINING PARALLEL")
+    stm_trainingp.add_argument("--nodes", default=1, type=int,
+                               help="number of nodes for distributed training, default 1")
+    stm_trainingp.add_argument("--ngpus_per_node", default=2, type=int,
+                               help="number of GPUs per node for distributed training, default 2")
+    stm_trainingp.add_argument("--dist-url", default="tcp://127.0.0.1:12315", type=str,
+                               help="url used to set up distributed training")
+    stm_trainingp.add_argument("--node_rank", default=0, type=int,
+                               help="node rank for distributed training, default 0")
+    stm_trainingp.add_argument("--epoch_sync", action="store_true", default=False,
+                               help="if sync model params of gpu0 to other local gpus after per epoch")
+
+    sub_trainm.set_defaults(func=main_trainm)
+
+    # # sub_denoise =====================================================================================
+    # sd_input = sub_denoise.add_argument_group("INPUT")
+    # sd_input.add_argument('--train_file', type=str, required=True, help="file containing (combined positive and "
+    #                                                                     "negative) samples for training. better been "
+    #                                                                     "balanced in kmer level.")
+    # sd_input.add_argument('--job_id', type=str, required=False, default="djob", help="job uid")
+    #
+    # sd_denoise = sub_denoise.add_argument_group("DENOISE")
+    # sd_denoise.add_argument('--is_filter_fn', type=str, default="yes", required=False,
+    #                         help="is filter false negative samples, 'yes' or 'no', default yes")
+    # sd_denoise.add_argument('--iterations', type=int, default=10, required=False)
+    # sd_denoise.add_argument('--rounds', type=int, default=3, required=False)
+    # sd_denoise.add_argument("--score_cf", type=float, default=0.2,
+    #                         required=False,
+    #                         help="score cutoff to keep high quality (which prob>=score_cf) positive samples. "
+    #                              "(0, 0.5], default 0.2")
+    # sd_denoise.add_argument("--kept_ratio", type=float, default=0.99,
+    #                         required=False,
+    #                         help="kept ratio of samples, to end denoise process. default 0.99")
+    # sd_denoise.add_argument("--fst_iter_prob", action="store_true", default=False,
+    #                         help="if output probs of samples after 1st iteration")
+    #
+    # sd_train = sub_denoise.add_argument_group("TRAIN MODEL_HYPER")
+    # # model param
+    # sd_train.add_argument('--model_type', type=str, default="attbigru2s",
+    #                       choices=["attbilstm2s", "attbigru2s"],
+    #                       required=False,
+    #                       help="type of model to use, 'attbilstm2s', 'attbigru2s', "
+    #                            "default: attbigru2s")
+    # sd_train.add_argument('--seq_len', type=int, default=21, required=False,
+    #                       help="len of kmer. default 21")
+    # sd_train.add_argument('--is_npass', type=str, default="yes", required=False,
+    #                       help="if using num_pass features, yes or no, default yes")
+    # sd_train.add_argument('--is_qual', type=str, default="no", required=False,
+    #                       help="if using base_quality features, yes or no, default no")
+    # sd_train.add_argument('--is_map', type=str, default="no", required=False,
+    #                       help="if using mapping features, yes or no, default no")
+    # sd_train.add_argument('--is_stds', type=str, default="no", required=False,
+    #                       help="if using std features, yes or no, default no")
+    # sd_train.add_argument('--class_num', type=int, default=2, required=False)
+    # sd_train.add_argument('--dropout_rate', type=float, default=0.5, required=False)
+    #
+    # # BiRNN model param
+    # sd_train.add_argument('--n_vocab', type=int, default=16, required=False,
+    #                       help="base_seq vocab_size (15 base kinds from iupac)")
+    # sd_train.add_argument('--n_embed', type=int, default=4, required=False,
+    #                       help="base_seq embedding_size")
+    # sd_train.add_argument('--layer_rnn', type=int, default=3,
+    #                       required=False, help="BiRNN layer num, default 3")
+    # sd_train.add_argument('--hid_rnn', type=int, default=256, required=False,
+    #                       help="BiRNN hidden_size for combined feature")
+    #
+    # sd_training = sub_denoise.add_argument_group("TRAINING")
+    # # model training
+    # sd_training.add_argument('--optim_type', type=str, default="Adam", choices=["Adam", "RMSprop", "SGD",
+    #                                                                             "Ranger"],
+    #                          required=False, help="type of optimizer to use, 'Adam' or 'SGD' or 'RMSprop' "
+    #                                               "or 'Ranger', default Adam")
+    # sd_training.add_argument('--batch_size', type=int, default=512, required=False)
+    # # sd_training.add_argument('--lr_scheduler', type=str, default='StepLR', required=False,
+    # #                          choices=["StepLR", "ReduceLROnPlateau"],
+    # #                          help="StepLR or ReduceLROnPlateau, default StepLR")
+    # sd_training.add_argument('--lr', type=float, default=0.001, required=False,
+    #                          help="default 0.001")
+    # sd_training.add_argument('--lr_decay', type=float, default=0.1, required=False,
+    #                          help="default 0.1")
+    # sd_training.add_argument('--lr_decay_step', type=int, default=1, required=False,
+    #                          help="effective in StepLR. default 1")
+    # # sd_training.add_argument('--lr_patience', type=int, default=0, required=False,
+    # #                          help="effective in ReduceLROnPlateau. default 0")
+    # # sd_training.add_argument('--lr_mode_strategy', type=str, default="last", required=False,
+    # #                          choices=["last", "mean", "max"],
+    # #                          help="effective in ReduceLROnPlateau. last, mean, or max, default last")
+    # sd_training.add_argument("--epoch_num", action="store", default=2, type=int,
+    #                          required=False, help="epoch num, default 2")
+    # sd_training.add_argument('--pos_weight', type=float, default=1.0, required=False)
+    # sd_training.add_argument('--step_interval', type=int, default=500, required=False)
+    # sd_training.add_argument('--dl_num_workers', type=int, default=0, required=False,
+    #                          help="default 0")
+    # sd_training.add_argument('--dl_offsets', action="store_true", default=False, required=False,
+    #                          help="use file offsets loader")
+    #
+    # sd_training.add_argument('--init_model', type=str, default=None, required=False,
+    #                          help="file path of pre-trained model parameters to load before training")
+    # sd_training.add_argument('--tseed', type=int, default=1234,
+    #                          help='random seed for pytorch')
+    #
+    # sub_denoise.set_defaults(func=main_denoise)
+
     args = parser.parse_args()
     if hasattr(args, 'func'):
         args.func(args)
